@@ -2,16 +2,22 @@
 See the README and class docstrings for more info.
 """
 import getpass
+import json
 import os
 from collections import defaultdict
+from datetime import datetime
 from decimal import Decimal, ROUND_HALF_UP
+from importlib.metadata import version
 from requests.exceptions import MissingSchema
+from urllib.error import URLError
 
 import altair as alt
 import click
 import pandas as pd
+from appdirs import user_data_dir
 from canvasapi import Canvas
 from canvasapi.exceptions import InvalidAccessToken, Unauthorized
+from luddite import get_version_pypi
 # Using https://github.com/biqqles/dataclassy instead of dataclasses from
 # stdlibto allow for dataclass inheritance when there are default values. Could
 # use a custom init but it gets messy and the advantage of using dataclasses is
@@ -21,6 +27,9 @@ from dataclassy import dataclass
 from . import __version__
 
 
+# Using this as a first try, might have to use an external file instead
+alt.data_transformers.disable_max_rows()
+
 @click.group()
 @click.version_option(version=__version__, prog_name='canvascli')
 def cli():
@@ -39,6 +48,59 @@ def cli():
     If you don't want to be prompted for your Canvas API token,
     you can save it to an environmental variable named CANVAS_PAT.
     """
+    # Since canvascli might be useful for people who don't use Python
+    # frequently, I have chosen to always print the program version and to
+    # occasionally check for newer versions of canvascli online and to note if
+    # one is found.
+    canvascli_version = version('canvascli')
+    click.echo(f'\ncanvascli version {canvascli_version}')
+
+    if 'canvascli_prevent_update' in os.environ:
+        pass
+    else:
+        data_dir = user_data_dir('canvascli')
+        data_file = os.path.join(data_dir, 'data.json')
+        try:
+            # If the config file exists
+            with open(data_file) as f:
+                stored_data = json.load(f)
+        except (FileNotFoundError, OSError, PermissionError):
+            # Generate a default config when no file is found
+            stored_data = {}
+            stored_data['last_update_check'] = '2001-01-01'
+
+        last_update_check = datetime.strptime(
+            stored_data['last_update_check'],
+            '%Y-%m-%d'
+        )
+        # Run a check online and update the config file if the last check was a long time ago
+        if (datetime.now() - last_update_check) > pd.Timedelta(weeks=4):
+            try:
+                latest_version = get_version_pypi("canvascli")
+                # Update `last_update_check` in the config if the update check is succesful
+                if canvascli_version == latest_version:
+                    os.makedirs(data_dir, exist_ok=True)
+                    with open(data_file, 'w', encoding='utf-8') as f:
+                        json.dump(
+                            {'last_update_check': datetime.now().strftime('%Y-%m-%d')},
+                            f,
+                            ensure_ascii=False,
+                            indent=4
+                        )
+                    # If no update was found, don't check again untill the env variables are cleared
+                    # This avoids reading the settings file unnecessarily
+                    # since the result will likely be the same in the same session anyways.
+                    os.environ['canvascli_prevent_update'] = 'True'
+                else:
+                    click.secho('\nNOTE', fg='yellow', bold=True)
+                    click.echo(
+                        f'A new version ({latest_version}) is available'
+                        '\nRead the release notes at https://github.com/joelostblom/canvascli/releases'
+                        '\nand download it via `python -m pip install -U canvascli`.'
+                    )
+            except URLError:
+                # If there is no connectivity
+                pass
     return
 
 
@@ -109,9 +171,12 @@ def prepare_fsc_grades(course_id, filename, api_url, student_status,
     fsc_grades.get_canvas_grades()
     fsc_grades.drop_student_entries()
     fsc_grades.convert_grades_to_fsc_format()
-    fsc_grades.save_fsc_grades_to_file()
-    fsc_grades.plot_fsc_grade_distribution()
-    fsc_grades.show_manual_grade_entry_note()
+    if fsc_grades.fsc_grades.empty:
+        click.echo('Did not find any assigned grades, exiting.')
+    else:
+        fsc_grades.save_fsc_grades_to_file()
+        fsc_grades.plot_fsc_grade_distribution()
+        fsc_grades.show_manual_grade_entry_note()
     return
 
 
@@ -121,7 +186,11 @@ def prepare_fsc_grades(course_id, filename, api_url, student_status,
 @click.option('--filter', 'filter_', default='', help='Only show courses including'
               ' this string (case insensitve). Useful to narrow down the displayed'
               ' results. Default: "" (matches all courses)')
-def show_courses(api_url, filter_):
+@click.option('--start-date', 'start_date_', default=None, help='Show courses'
+              ' starting at or after this date. Courses without a start date are'
+              ' always shown. Format "YYYY-MM-DD". Default: Show all courses'
+              ' starting within a year from today.')
+def show_courses(api_url, filter_, start_date_):
     """Show courses accessible by the given API token.
 
     Examples:
@@ -132,7 +201,7 @@ def show_courses(api_url, filter_):
         # Show only courses including the string "541"
         canvascli show-courses --filter 541
     """
-    accessible_courses = AccessibleCourses(api_url, filter_)
+    accessible_courses = AccessibleCourses(api_url, filter_, start_date_)
     accessible_courses.connect_to_canvas()
     accessible_courses.download_courses()
     accessible_courses.filter_and_show_courses()
@@ -178,14 +247,19 @@ class AccessibleCourses(CanvasConnection):
     """Show all courses accessible from a specific API token."""
     api_url: str
     filter_: str
+    start_date_: str
 
     def download_courses(self):
         """Download all courses accessible from a specific API token."""
         self.courses = defaultdict(list)
         try:
             for course in self.canvas.get_courses():
-                self.courses['id'].append(course.id)
-                self.courses['name'].append(course.name)
+
+                # Return a default value ('N/A') if the attribute is not found
+                self.courses['id'].append(getattr(course, 'id', 'N/A'))
+                self.courses['name'].append(getattr(course, 'name', 'N/A'))
+                self.courses['end_date'].append(getattr(course, 'end_at', 'N/A'))
+                self.courses['start_at'].append(getattr(course, 'start_at', 'N/A'))
         # Show common exceptions in a way that is easy to understand
         except MissingSchema:
             raise SystemExit(self.invalid_canvas_url_msg)
@@ -195,11 +269,32 @@ class AccessibleCourses(CanvasConnection):
 
     def filter_and_show_courses(self):
         """Filter and show downloaded courses."""
+        # The default is to include courses starting in the past year
+        start_date = pd.to_datetime(
+            self.start_date_ or datetime.now() - pd.DateOffset(months=12),
+            utc=True
+        ).date()
         click.echo("Your API token has access to the following courses:\n")
         click.echo(
-            pd.DataFrame(self.courses)
-            .query('name.str.contains(@self.filter_, case=False)', engine='python')
-            .to_markdown(index=False))
+            pd.DataFrame(
+                self.courses
+            ).assign(
+                start_at = lambda df: pd.to_datetime(df['start_at']).dt.date
+            ).query(
+                'start_at > @start_date'
+            ).query(
+                'name.str.contains(@self.filter_, case=False)',
+                engine='python'
+            ).sort_values(
+                'start_at'
+            ).drop(
+                columns=['end_date']
+            ).rename(
+                columns={'id': 'ID', 'name': 'Name', 'start_at': 'Start Date'}
+            ).to_markdown(
+                index=False
+            )
+        )
         return
 
 
@@ -243,31 +338,74 @@ class FscGrades(CanvasConnection):
         """Download grades from a canvas course."""
         click.echo('Downloading student grades...')
         enrollments = self.course.get_enrollments(
-            type=['StudentEnrollment'], state=[self.student_status])
+            type=['StudentEnrollment'], state=[self.student_status]
+        )
         canvas_grades = defaultdict(list)
-        overridden_grades = []
         for enrollment in enrollments:
             canvas_grades['Student Number'].append(enrollment.user['sis_user_id'])
             surname, preferred_name = enrollment.user['sortable_name'].split(', ')
             canvas_grades['Surname'].append(surname)
             canvas_grades['Preferred Name'].append(preferred_name)
+
+            # A warning message is later displayed for these students
             if 'override_score' in enrollment.grades:
-                overridden_grades.append(
-                    f"{preferred_name:<12} {surname:<12}"
-                    f" {enrollment.grades['final_score']:<4} ->"
-                    f" {enrollment.grades['override_score']}")
                 canvas_grades['Percent Grade'].append(enrollment.grades['override_score'])
+                canvas_grades['override_final_score'].append(enrollment.grades['final_score'])
             else:
                 canvas_grades['Percent Grade'].append(enrollment.grades['final_score'])
+                canvas_grades['override_final_score'].append(0)
+
+            # A warning message is later displayed for these students
             if 'unposted_final_score' in enrollment.grades:
+                canvas_grades['Unposted Percent Grade'].append(enrollment.grades['unposted_final_score'])
                 if enrollment.grades['unposted_final_score'] != enrollment.grades['final_score']:
-                    click.echo(
-                        'There are unposted assignments that would change the final score of '
-                        + f'{preferred_name} {surname}. Please post all assignments on Canvas.')
+                    canvas_grades['different_unposted_score'].append(True)
+                else:
+                    canvas_grades['different_unposted_score'].append(False)
+
         self.canvas_grades = pd.DataFrame(canvas_grades)
-        if len(overridden_grades) > 0:
-            click.echo('Students with manual Canvas override of their final grade:')
-            [click.echo(grade) for grade in overridden_grades]
+        different_unposted_score = self.canvas_grades.pop('different_unposted_score')
+        override_final_score = self.canvas_grades.pop('override_final_score')
+
+        # Display a note that some student grade are manually overridden
+        if override_final_score.sum() > 0:
+            click.secho('\nNOTE', fg='yellow', bold=True)
+            click.echo(
+                'You have used the "Overide" column on Canvas'
+                '\nto change the final score for the following students:\n'
+            )
+            click.echo(
+                self.canvas_grades
+                    .query(
+                        '@override_final_score > 0'
+                    ).assign(
+                        **{'Percent Grade Before Override': override_final_score}
+                    ).drop(
+                        columns='Unposted Percent Grade'
+                    ).to_markdown(
+                        index=False
+                    )
+                )
+            click.echo()
+
+        # Warn about students with unposted grades that change their final scores
+        if different_unposted_score.sum() > 0:
+            click.secho('\nWARNING', fg='red', bold=True)
+            click.echo(
+                'Remember to post all assignments on Canvas'
+                '\nbefore creating the CSV-file to upload to the FSC.'
+                '\nThere are currently unposted Canvas assignments'
+                '\nthat would change the final score of the following students:\n'
+            )
+            click.echo(
+                self.canvas_grades
+                    .query(
+                        '@different_unposted_score == True'
+                    )
+                    .to_markdown(
+                        index=False
+                    )
+                )
             click.echo()
         return
 
@@ -276,9 +414,9 @@ class FscGrades(CanvasConnection):
         # Drop students under the grade thresholds
         # Test accounts and students who dropped the course often have a grade of zero
         dropped_students = self.canvas_grades.query(
-            '`Percent Grade` <= @self.drop_threshold')
+            '`Unposted Percent Grade` <= @self.drop_threshold')
         self.canvas_grades = self.canvas_grades.query(
-            '`Percent Grade` > @self.drop_threshold').copy()
+            '`Unposted Percent Grade` > @self.drop_threshold').copy()
 
         # Drop students that have missing info in any field
         # These are also printed so that it is clear to the user what has happened
@@ -299,8 +437,13 @@ class FscGrades(CanvasConnection):
 
         # Display the dropped students so the user can catch errors easily
         if dropped_students.shape[0] > 0:
-            click.echo(f'Dropping {dropped_students.shape[0]} student(s) with'
-                       f' missing information or a grade <= {self.drop_threshold}:\n')
+            click.secho('\nNOTE', fg='yellow', bold=True)
+            click.echo(
+                f'Dropping {dropped_students.shape[0]}'
+                f" {'students' if dropped_students.shape[0] > 1 else 'student'}"
+                f' with missing information, a grade <= {self.drop_threshold},'
+                '\nor that was explicitly dropped by student number:\n'
+            )
             click.echo(dropped_students.to_markdown(index=False))
             click.echo()
         return
@@ -326,62 +469,201 @@ class FscGrades(CanvasConnection):
         # Add FSC info to the dataframe, standing and standing reason are
         # blank by default and filled out manually when needed
         self.fsc_grades = self.canvas_grades.copy()
-        fsc_fields = ['Campus', 'Course', 'Section', 'Session', 'Subject',
+        additional_fsc_fields = ['Campus', 'Course', 'Section', 'Session', 'Subject',
                       'Standing', 'Standing Reason']
         # Remove the session number which is only present on Canvas but not FSC
         self.session = self.session[:-1]
-        self.fsc_grades[fsc_fields] = (
+        self.fsc_grades[additional_fsc_fields] = (
             self.campus, self.course, self.section, self.session, self.subject, '', '')
-        # Reorder columns to match the required FSC format
-        self.fsc_grades = self.fsc_grades[[
-            'Session', 'Campus', 'Student Number', 'Subject', 'Course', 'Section',
-            'Surname', 'Preferred Name', 'Standing', 'Standing Reason', 'Percent Grade']]
+
+        # Round to whole percentage format since FSC requires that
         # Using Decimal to always round up .5 instead of rounding to even,
         # which is the default in numpy but could seem unfair to individual students.
         # The Decimal type is not json serializable (issue for altair) so changing to int.
+        self.fsc_grades['Exact Percent Grade'] = self.fsc_grades['Percent Grade']
         self.fsc_grades['Percent Grade'] = self.fsc_grades['Percent Grade'].apply(
             lambda x: Decimal(x).quantize(0, rounding=ROUND_HALF_UP)).astype(int)
+
+        self.fsc_grades['Unposted Exact Percent Grade'] = self.fsc_grades['Unposted Percent Grade']
+        self.fsc_grades['Unposted Percent Grade'] = self.fsc_grades['Unposted Percent Grade'].apply(
+            lambda x: Decimal(x).quantize(0, rounding=ROUND_HALF_UP)).astype(int)
+
         # Cap grades at 100
         self.fsc_grades.loc[self.fsc_grades['Percent Grade'] > 100, 'Percent Grade'] = 100
+
         return
 
     def save_fsc_grades_to_file(self):
         """Write a CSV file that can be uploaded to FSC."""
-        self.fsc_grades.to_csv(self.filename + '.csv', index=False)
+        # Reorder columns to match the required FSC format
+        self.fsc_grades[[
+            'Session', 'Campus', 'Student Number', 'Subject', 'Course', 'Section',
+            'Surname', 'Preferred Name', 'Standing', 'Standing Reason', 'Percent Grade'
+        ]].to_csv(
+            self.filename + '.csv',
+            index=False
+        )
         click.echo(f'Grades saved to {self.filename}.csv.')
         return
 
     def plot_fsc_grade_distribution(self):
-        """Create a histogram and interactive strip chart for the grade distribution"""
-        hist = (
-            alt.Chart(self.fsc_grades, height=200).mark_bar().encode(
-                alt.X('Percent Grade', bin=alt.Bin(maxbins=15), title='', axis=alt.Axis(labels=False)),
-                alt.Y('count()', title='Number of students')))
-        strip = (
-            alt.Chart(self.fsc_grades, height=60)
-            .mark_point(size=20)
-            .transform_calculate(
-                # Generate Gaussian jitter with a Box-Muller transform
-                jitter='sqrt(-2*log(random()))*cos(2*PI*random())',
-                Name='datum["Preferred Name"] + " " + datum["Surname"]')
-            .encode(
-                alt.X('Percent Grade', scale=alt.Scale(zero=False, nice=False, padding=5)),
-                alt.Y('jitter:Q', scale=alt.Scale(padding=2), axis=alt.Axis(
-                    domain=False, title='', labels=False, ticks=False, grid=False)),
-                alt.Tooltip(['Name:N', 'Student Number', 'Percent Grade']))
-            .interactive())
+        # Prepare dataframe for filtering via Altair selection elements
+        # First the rounded and raw scores are melted together separately for posted and unposted scores
+        # Then they are merged into one frame and the posted and unposted score are melted together
+        self.fsc_grades_for_viz = pd.merge(
+            # Frame 1
+            self.fsc_grades.rename(
+                columns={
+                    'Unposted Percent Grade': 'FSC Rounded',
+                    'Unposted Exact Percent Grade': 'Exact'
+                }
+            # Combine the rounded and raw *unposted* scores 
+            ).melt(
+                id_vars=['Preferred Name', 'Surname', 'Student Number'],
+                value_vars=['FSC Rounded', 'Exact'],
+                value_name='Unposted Grade',
+                var_name='Percent Type'
+            ),
+            # Frame 2
+            self.fsc_grades.rename(
+                columns={
+                    'Percent Grade': 'FSC Rounded',
+                    'Exact Percent Grade': 'Exact'
+                }
+            # Combine the rounded and raw *posted* scores 
+            ).melt(
+                id_vars=['Preferred Name', 'Surname', 'Student Number'],
+                value_vars=['FSC Rounded', 'Exact'],
+                value_name='Posted Grade',
+                var_name='Percent Type'
+            )
+        # Combine the posted and unposted scores
+        ).melt(
+            id_vars=['Preferred Name', 'Surname', 'Student Number', 'Percent Type'],
+            value_vars=['Unposted Grade', 'Posted Grade'],
+            value_name='Percent Grade',
+            var_name='Grade Status'
+        )
+
+        # Set up selection elements
+        grade_status_dropdown = alt.binding_select(
+            options=['Posted Grade', 'Unposted Grade'],
+            name='Grade Status '
+        )
+        grade_status_selection = alt.selection_single(
+            fields=['Grade Status'],
+            bind=grade_status_dropdown,
+            init={'Grade Status': 'Posted Grade'}
+        )
+        percent_type_dropdown = alt.binding_select(
+            options=['FSC Rounded', 'Exact'],
+            name='Percent Type '
+        )
+        percent_type_selection = alt.selection_single(
+            fields=['Percent Type'],
+            bind=percent_type_dropdown,
+            init={'Percent Type': 'FSC Rounded'}
+        )
+
+        # Plot distribution
+        hist = alt.Chart(self.fsc_grades_for_viz, height=200).mark_bar().encode(
+            alt.X('Percent Grade', bin=alt.Bin(maxbins=15), title='', axis=alt.Axis(labels=False)),
+            alt.Y('count()', title='Number of students')
+        )
+
+        # Plot all observations
+        strip = alt.Chart(self.fsc_grades_for_viz, height=70).mark_point(
+            size=20
+        ).transform_calculate(
+            # Generate Gaussian jitter with a Box-Muller transform
+            jitter='sqrt(-2*log(random()))*cos(2*PI*random())',
+            Name='datum["Preferred Name"] + " " + datum["Surname"]'
+        ).encode(
+            alt.X('Percent Grade',
+                title='Percent Grade',
+                scale=alt.Scale(
+                    zero=False,
+                    nice=False,
+                    padding=5
+                )
+            ),
+            alt.Y('jitter:Q',
+                scale=alt.Scale(padding=2),
+                axis=alt.Axis(
+                    domain=False,
+                    title='',
+                    labels=False,
+                    ticks=False,
+                    grid=False
+                )
+            ),
+            alt.Tooltip(['Name:N', 'Student Number', 'Percent Grade'])
+        ).interactive()
+
+        # Plot central tendencies
+        central_tendencies = alt.Chart(
+                self.fsc_grades_for_viz
+            ).transform_aggregate(
+                Mean='mean(Percent Grade)',
+                Median='median(Percent Grade)',
+            ).transform_fold(
+                fold=['Mean', 'Median'],
+                as_=['Type', 'Percent Grade']
+            ).transform_calculate(
+                y='-3.05',
+            ).mark_point(
+            size=65,
+            shape='diamond'
+        ).encode(
+            x='Percent Grade:Q',
+            y='y:Q',
+            color=alt.Color(
+                'Type:N',
+                title='',
+                scale=alt.Scale(range=['coral', 'rebeccapurple']),
+                legend=alt.Legend(legendY=225, legendX=415, orient='none')
+            ),
+            tooltip=['Type:N', alt.Tooltip('Percent Grade:Q', format='.3g')]
+        )
+
+        # Add instructions
         title = alt.TitleParams(
             text=f'Grade distribution {self.subject} {self.course}',
             subtitle=[
                 'Hover over the points to see the student name and grade.',
                 'Zoom with the mouse wheel and double click to reset the view.',
-                'Click the three dots button to the right to save the plot.'])
+                'Click the three dots button to the right to save the plot.']
+        )
+
+        # Concatenate, add filters, and save the chart
         chart_filename = self.filename + '.html'
-        (alt.vconcat(hist, strip, spacing=0)
-            .properties(title=title)
-            .resolve_scale(x='shared')
-            .configure_view(strokeWidth=0)
-            .save(chart_filename))
+        alt.vconcat(
+            hist,
+            # strip on top so that individual observations are always visible
+            central_tendencies + strip,
+            spacing=0
+        ).properties(
+            title=title
+        ).resolve_scale(
+            x='shared'
+        ).configure_view(
+            strokeWidth=0
+        ).transform_filter(
+                percent_type_selection & grade_status_selection
+        ).add_selection(
+            percent_type_selection, grade_status_selection
+        ).save(
+            chart_filename
+        )
+        # The label names are styled in serif by default for some reason
+        with open(chart_filename, 'a') as chart_file:
+            chart_file.write(
+                '\n<style>'
+                '\n    form.vega-bindings {'
+                '\n        font-family: sans-serif;'
+                '\n    }'
+                '\n</style>'
+            )
         click.echo(f'Grade distribution chart saved to {chart_filename}.')
         if self.open_chart or self.open_chart is None and click.confirm(
                 'Open grade distribution chart?', default=True):
@@ -390,12 +672,14 @@ class FscGrades(CanvasConnection):
 
     def show_manual_grade_entry_note(self):
         """Show disclaimer and note about manual grade entry of student standings."""
+        click.secho('\nNOTE', fg='yellow', bold=True)
         click.echo(
-            '\nNote:\nThe saved CSV file should automatically be correctly formatted,'
-            '\nbut it is a good idea to double check in case there are unexpected changes'
-            '\nto how UBC inputs course info on Canvas.'
-            '\nIf you have students that did not take the final exam'
-            '\nor with a thesis in progress, you will need to enter this info manually.'
-            '\nPlease see https://facultystaff.students.ubc.ca/files/FSCUserGuide.pdf'
-            '\nunder "Acceptable Values for Grades Entry" for how to modify the CSV file.')
+            '1. Make sure to check the output above for any warning and notes.'
+            '\n2. The saved CSV file should automatically be correctly formatted,'
+            '\n   but it is ' + click.style('your responsibility', bold=True) + ' to double check in case there are unexpected changes'
+            '\n   to how UBC inputs course info on Canvas.'
+            '\n3. If you have students that did not take the final exam'
+            '\n   or with a thesis in progress, you will need to enter this info manually.'
+            '\n   Please see https://facultystaff.students.ubc.ca/files/FSCUserGuide.pdf'
+            '\n   under "Acceptable Values for Grades Entry" for how to modify the CSV file.')
         return
